import os
import shutil
import six
import functools

import pytest


class BaseTest(object):
    root_path = os.path.split(os.path.abspath(__name__))[0]


def rmdir(path):
    if os.path.exists(path):
        shutil.rmtree(path)


def chdir(path):
    def wrapper(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            cwd = os.getcwd()
            if not os.path.exists(path):
                os.makedirs(path)
            os.chdir(path)
            func(*args, **kwargs)
            os.chdir(cwd)
        return inner
    return wrapper


python2_only = pytest.mark.skipif(not six.PY2, reason='only support Python2')
python3_only = pytest.mark.skipif(not six.PY3, reason='only support Python3')



# -*- encoding: utf-8 -*-
from flask_sqlalchemy import model

from marshmallow import fields
from webargs.fields import DelimitedList

from .utils import ParamsDict, dict2object


#: Base params for list view func.
PageParams = ParamsDict(
    page=fields.Int(missing=1, required=False),
    page_size=fields.Int(missing=10, required=False),
    order_by=DelimitedList(
        fields.String(missing='id'), required=False, missing=['id']),
)
"""Base params for list view func which contains ``page``、``page_size``、\
   ``order_by`` params.

    Example::

        @use_kwargs(PageParams)
        def list_users(page, page_size, order_by):
            pass
"""


def pagination(obj, page, page_size, order_by='id', query_exp=None):
    """A pagination for sqlalchemy query.

    Args:
        obj (db.Model): Model class like User.
        page (int): Page index.
        page_size (int): Row's count per page.
        order_by (str, list): Example: 'id'、['-id', 'column_name'].
        query_exp (flask_sqlalchemy.BaseQuery): Query like \
            ``User.query.filter_by(id=1)``.

    Returns:
        class: Class that contains ``items``、``page``、``page_size`` and \
            ``total`` fileds.
    """
    if not isinstance(obj, model.DefaultMeta):
        raise Exception('first arg obj must be model.')

    if not isinstance(order_by, list):
        order_by = [order_by]

    columns = {i.name for i in obj.__table__.columns}
    diff = {c.lstrip('-') for c in order_by} - columns
    if diff:
        raise Exception('columns {} not exist in {} model'.format(diff, obj))

    order_exp = []
    for column in order_by:
        if column.startswith('-'):
            order_exp.append(getattr(obj, column.lstrip('-')).desc())
        else:
            order_exp.append(getattr(obj, column))

    qexp = query_exp or getattr(obj, 'query')

    items = qexp.order_by(*order_exp).paginate(
        page, page_size, error_out=False)

    return dict2object({
        'items': items.items, 'page': page, 'page_size': page_size,
        'total': items.total,
    })



import pytest
from .run import app as tapp


@pytest.fixture(scope='session')
def app(request):
    ctx = tapp.app_context()
    ctx.push()

    def teardown():
        ctx.pop()

    request.addfinalizer(teardown)
    return tapp


@pytest.fixture(scope='session')
def client(app, request):
    return app.test_client()



import os
from flask import Flask

from .exts import db, ma, hobbit
from .views import bp

ROOT_PATH = os.path.split(os.path.abspath(__file__))[0]


class ConfigClass(object):
    SECRET_KEY = 'test secret key'
    SQLALCHEMY_DATABASE_URI = 'sqlite:///{}/tst_app.sqlite'.format(ROOT_PATH)
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    TESTING = True


def register_extensions(app):
    db.init_app(app)
    ma.init_app(app)
    hobbit.init_app(app)


def register_blueprints(app):
    app.register_blueprint(bp)


def init_app(config=ConfigClass):
    app = Flask(__name__)
    app.config.from_object(config)

    register_extensions(app)
    register_blueprints(app)

    return app


app = init_app()



# -*- encoding: utf-8 -*-
import pytest

from hobbit_core.flask_hobbit import utils

from . import BaseTest, python2_only, python3_only


class TestUtils(BaseTest):

    def test_dict2object(self):
        obj = utils.dict2object({'a': 2, 'c': 3})
        assert obj.a == 2
        assert obj.c == 3

        # test setattr
        obj.a = 4
        assert obj.a == 4

        # test getattr
        with pytest.raises(AttributeError):
            obj.b

    def test_secure_filename(self):
        filenames = (
            u'哈哈.zip', '../../../etc/passwd', 'My cool movie.mov',
            '__filename__', 'foo$&^*)bar',
            u'i contain cool \xfcml\xe4uts.txt',
        )
        excepted = (
            u'哈哈.zip', 'etc_passwd', 'My_cool_movie.mov',
            'filename', 'foobar',
            'i_contain_cool_umlauts.txt',
        )
        for i, filename in enumerate(filenames):
            assert utils.secure_filename(filename) == excepted[i]

    @python2_only
    def test_secure_filename_py2(self):
        with pytest.raises(
                Exception, message="filename must be <type 'unicode'>"):
            assert utils.secure_filename(
                'i contain cool \xfcml\xe4uts.txt') == \
                'i_contain_cool_umlauts.txt'

    @python3_only
    def test_secure_filename_py3(self):
        assert utils.secure_filename(
            'i contain cool \xfcml\xe4uts.txt') == \
            'i_contain_cool_umlauts.txt'

    def test_use_kwargs(self, client):
        payload = {'username': 'username', 'email': 'email'}
        resp = client.post('/use_kwargs_with_partial/', json=payload)
        assert resp.json == payload

        resp = client.post('/use_kwargs_without_partial/', json=payload)
        assert resp.json == payload

        payload = {'username': 'username'}
        resp = client.post('/use_kwargs_with_partial/', json=payload)
        assert resp.json == payload

        resp = client.post('/use_kwargs_without_partial/', json=payload)
        assert resp.json == {'username': 'username', 'email': None}



# -*- encoding: utf-8 -*-
import traceback

from sqlalchemy.orm import exc as orm_exc

from .response import Result, ServerErrorResult, gen_response, RESP_MSGS


class ErrHandler(object):
    """Base error handler that catch all exceptions. Be sure response is::

        {
            "code": "404",  # error code, default is http status code, \
you can change it
            "message": "Not found",  # for alert in web page
            "detail": "id number field length must be 18",  # for debug
        }

    Examples::

        app.register_error_handler(Exception, ErrHandler.handler)
    """

    @classmethod
    def handler_werkzeug_exceptions(cls, e):
        return Result(gen_response(
            e.code, RESP_MSGS.get(e.code, e.name),
            None if not hasattr(e, 'data') else e.data['messages']),
            status=e.code)

    @classmethod
    def handler_sqlalchemy_orm_exc(cls, e):
        code, message, detail = 500, RESP_MSGS[500], repr(e)

        if isinstance(e, orm_exc.NoResultFound):
            code, message, detail = 404, '源数据未找到', repr(e)

        return Result(gen_response(code, message, detail), status=code)

    @classmethod
    def handler_others(cls, e):
        traceback.print_exc()
        return ServerErrorResult(500, detail=repr(e))

    @classmethod
    def handler(cls, e):
        exc = 'others' if not hasattr(e, '__module__') else \
            e.__module__.replace('.', '_')
        return getattr(cls, 'handler_{}'.format(exc), cls.handler_others)(e)



from contextlib import contextmanager
import os

import click
from jinja2 import Environment, FileSystemLoader, Template
from . import echo

SUFFIX = '.jinja2'
EXAMPLE_SUFFIX = 'example.py.jinja2'


@contextmanager
def chdir(dist):
    cwd = os.getcwd()
    echo('mkdir\t{}', (dist, ))
    # exist_ok py3 only
    if not os.path.exists(dist):
        os.makedirs(dist)
    os.chdir(dist)
    yield dist
    os.chdir(cwd)


@click.pass_context
def render_project(ctx, dist, tpl_path):
    example = ctx.obj['EXAMPLE']
    jinjia_env = Environment(loader=FileSystemLoader(tpl_path))
    with chdir(dist):
        for fn in os.listdir(tpl_path):
            origin_path = os.path.join(tpl_path, fn)

            if os.path.isfile(origin_path) and not fn.endswith(SUFFIX):
                continue
            if not example and fn.endswith(EXAMPLE_SUFFIX):
                continue

            if os.path.isfile(origin_path):
                data = jinjia_env.get_template(fn).render(
                    ctx.obj['JINJIA_CONTEXT'])
                render_file(dist, fn[:-len(SUFFIX)], data)
                continue

            dir_name = Template(fn).render(ctx.obj['JINJIA_CONTEXT'])
            render_project(os.path.join(dist, dir_name),
                           os.path.join(tpl_path, fn))


@click.pass_context
def render_file(ctx, dist, fn, data):
    target = os.path.join(dist, fn)
    if os.path.isfile(fn) and not ctx.obj['FORCE']:
        echo('exists {}, ignore ...', (target, ))
        return

    echo('render\t{} ...', (target, ))

    with open(fn, 'w') as wf:
        wf.write(data)

    if fn.endswith('.sh'):
        os.chmod(fn, 0o755)



# -*- encoding: utf-8 -*-
from marshmallow import Schema, fields, post_load
from flask_marshmallow.sqla import ModelSchema


class ORMSchema(ModelSchema):
    """Base schema for ModelSchema. See `webargs/issues/126
    <https://github.com/sloria/webargs/issues/126>`_.

    Example::

        from hobbit_core.flask_hobbit.schemas import ORMSchema


        class UserSchema(ORMSchema):

            class Meta:
                model = User
                load_only = ('password')

    ``@use_kwargs(UserSchema())`` use in combination with ``load_only``::

        @bp.route('/users/', methods=['POST'])
        @use_kwargs(UserSchema())
        def create_user(username, password):
            pass
    """

    @post_load()
    def make_instance(self, data):
        return data


class SchemaMixin(object):
    """Add ``id``, ``created_at``, ``updated_at`` fields to schema,
    default ``dump_only=True``.

    Example::

        from marshmallow import Schema

        from hobbit_core.flask_hobbit.schemas import SchemaMixin

        class UserSchema(Schema, SchemaMixin):
            pass
    """

    id = fields.Int(dump_only=True)
    created_at = fields.DateTime('%Y-%m-%d %H:%M:%S', dump_only=True)
    updated_at = fields.DateTime('%Y-%m-%d %H:%M:%S', dump_only=True)

    class Meta:
        strict = True
        ordered = False
        dateformat = '%Y-%m-%d %H:%M:%S'


class PagedSchema(Schema):
    """Base schema for list api pagination.

    Example::

        from marshmallow import fields

        from hobbit_core.flask_hobbit.schemas import PagedSchema

        from . import models
        from .exts import ma


        class UserSchema(ma.ModelSchema):

            class Meta:
                model = models.User


        class PagedUserSchema(PagedSchema):
            items = fields.Nested('UserSchema', many=True)


        paged_user_schemas = PagedUserSchema()
    """

    total = fields.Int()
    page = fields.Int(missing=1, default=1)
    page_size = fields.Int(missing=10, default=10)

    class Meta:
        strict = True



import click

from . import bootstrap


class CLI(click.MultiCommand):

    def list_commands(self, ctx):
        return sorted(self.cmds.keys())

    def get_command(self, ctx, cmd_name):
        try:
            return self.cmds[cmd_name]
        except KeyError:
            raise click.UsageError(click.style(
                "cmd not exist: {}\nAvailable ones are: {}".format(
                    cmd_name, ', '.join(self.cmds),
                ), fg='red'))

    @property
    def cmds(self):
        return {func.name: func for func in bootstrap.CMDS}


@click.command(cls=CLI)
@click.version_option()
@click.option('--echo/--no-echo', default=False, help='Show cmd run msg.')
@click.pass_context
def main(ctx, echo):
    if ctx.obj is None:
        ctx.obj = dict()
    ctx.obj['ECHO'] = echo


if __name__ == '__main__':
    main(obj={})



# -*- encoding: utf-8 -*-
from collections import Mapping
import inspect
import os
import re
import six
from unicodedata import normalize

from flask import request
from marshmallow import Schema
from webargs.flaskparser import use_kwargs as base_use_kwargs, parser


class ParamsDict(dict):
    """Just available update func.

    Example::

        @use_kwargs(PageParams.update({...}))
        def list_users(page, page_size, order_by):
            pass

    """

    def update(self, other=None):
        """Update self by other Mapping and return self.
        """
        ret = self.copy()
        if other is not None:
            for k, v in other.items() if isinstance(other, Mapping) else other:
                ret[k] = v
        return ret


class dict2object(dict):
    """Dict to fake object that can use getattr.

    Examples::

        In [2]: obj = dict2object({'a': 2, 'c': 3})

        In [3]: obj.a
        Out[3]: 2

        In [4]: obj.c
        Out[4]: 3

    """

    def __getattr__(self, name):
        if name in self.keys():
            return self[name]
        raise AttributeError('object has no attribute {}'.format(name))

    def __setattr__(self, name, value):
        if not isinstance(name, six.string_types):
            raise TypeError('key must be string type.')
        self[name] = value


def secure_filename(filename):
    """Borrowed from werkzeug.utils.secure_filename.

    Pass it a filename and it will return a secure version of it. This
    filename can then safely be stored on a regular file system and passed
    to :func:`os.path.join`.

    On windows systems the function also makes sure that the file is not
    named after one of the special device files.

        >>> secure_filename(u'哈哈.zip')
        '哈哈.zip'
        >>> secure_filename('My cool movie.mov')
        'My_cool_movie.mov'
        >>> secure_filename('../../../etc/passwd')
        'etc_passwd'
        >>> secure_filename(u'i contain cool \xfcml\xe4uts.txt')
        'i_contain_cool_umlauts.txt'

    """
    if not isinstance(filename, six.text_type):
        try:
            filename = filename.decode('utf-8')
        except UnicodeDecodeError:
            raise Exception(
                'filename must be {}'.format(six.text_type))

    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, ' ')

    filename = '_'.join(filename.split())

    if isinstance(filename, six.text_type):
        filename = normalize('NFKD', filename).encode('utf-8')
        if not six.PY2:
            filename = filename.decode('utf-8')

    if six.PY2 and not isinstance(filename, six.text_type):
        filename = filename.decode('utf-8', 'replace')

    filename_strip_re = re.compile(u'[^A-Za-z0-9\u4e00-\u9fa5_.-]')
    filename = filename_strip_re.sub('', filename).strip('._')

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    windows_device_files = (
        'CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4', 'LPT1',
        'LPT2', 'LPT3', 'PRN', 'NUL',
    )
    if os.name == 'nt' and filename and \
            filename.split('.')[0].upper() in windows_device_files:
        filename = '_' + filename

    return filename


def _get_init_args(instance, base_class):
    """Get instance's __init__ args and it's value when __call__.
    """
    if six.PY2:
        getargspec = inspect.getargspec
    else:
        getargspec = inspect.getfullargspec

    argspec = getargspec(base_class.__init__)
    no_defaults = argspec.args[:-len(argspec.defaults)]
    has_defaults = argspec.args[-len(argspec.defaults):]

    kwargs = {k: getattr(instance, k) for k in no_defaults
              if k != 'self' and hasattr(instance, k)}
    kwargs.update({k: getattr(instance, k, argspec.defaults[i])
                   for i, k in enumerate(has_defaults)})

    assert len(kwargs) == len(argspec.args) - 1, 'exclude `self`'

    return kwargs


def use_kwargs(argmap, schema_kwargs=None, **kwargs):
    """For fix ``Schema(partial=True)`` not work when used with
    ``@webargs.flaskparser.use_kwargs``. More details ``see webargs.core``.

    Args:

        argmap (marshmallow.Schema,dict,callable): Either a
            `marshmallow.Schema`, `dict` of argname ->
            `marshmallow.fields.Field` pairs, or a callable that returns a
            `marshmallow.Schema` instance.
        schema_kwargs (dict): kwargs for argmap.

    Returns:
        dict: A dictionary of parsed arguments.

    """
    schema_kwargs = schema_kwargs or {}

    argmap = parser._get_schema(argmap, request)

    if not (argmap.partial or schema_kwargs.get('partial')):
        return base_use_kwargs(argmap, **kwargs)

    def factory(request):
        argmap_kwargs = _get_init_args(argmap, Schema)
        argmap_kwargs.update(schema_kwargs)

        # force set force_all=False
        only = parser.parse(argmap, request).keys()

        argmap_kwargs.update({
            'partial': True,
            'only': only or None,
            'context': {"request": request},
            'strict': True,
        })

        return argmap.__class__(**argmap_kwargs)

    return base_use_kwargs(factory, **kwargs)



import os

from click.testing import CliRunner

from hobbit_core.hobbit import main as hobbit

from . import BaseTest, rmdir, chdir


class TestHobbit(BaseTest):
    wkdir = os.path.abspath('hobbit-tox-test')

    def setup_method(self, method):
        rmdir(self.wkdir)

    def teardown_method(self, method):
        os.chdir(self.root_path)
        rmdir(self.wkdir)

    def test_hobbit_cmd(self):
        runner = CliRunner()

        result = runner.invoke(hobbit)
        assert result.exit_code == 0

        result = runner.invoke(hobbit, ['doesnotexistcmd'], obj={})
        assert 'Error: cmd not exist: doesnotexistcmd' in result.output

    @chdir(wkdir)
    def test_startproject_cmd_nodist(self):
        assert os.getcwd() == self.wkdir
        runner = CliRunner()

        result = runner.invoke(hobbit, ['--echo', 'startproject'], obj={})
        assert result.exit_code == 2
        assert 'Error: Missing option "-n" / "--name".' in result.output

        result = runner.invoke(
            hobbit, [
                '--echo', 'startproject', '-n', 'haha', '-f', '-p', '1024',
            ], obj={})
        assert result.exit_code == 0
        assert 'mkdir\t{}'.format(self.wkdir) in result.output
        assert 'render\t{}'.format(self.wkdir) in result.output
        assert os.path.exists(os.path.join(
            os.getcwd(), 'app', 'models', '__init__.py'))
        assert 'example.py' not in result.output

    def test_startproject_cmd_dist(self):
        runner = CliRunner()

        result = runner.invoke(
            hobbit, [
                '--echo', 'startproject', '-n', 'haha', '-f', '-d', self.wkdir,
                '-p', '1024',
            ], obj={})
        assert result.exit_code == 0
        assert 'mkdir\t{}'.format(self.wkdir) in result.output
        assert 'render\t{}'.format(self.wkdir) in result.output
        assert os.path.exists(os.path.join(
            self.wkdir, 'app', 'models', '__init__.py'))
        assert 'example.py' not in result.output

    @chdir(wkdir)
    def test_startproject_cmd_curdir(self):
        assert os.getcwd() == self.wkdir
        runner = CliRunner()

        result = runner.invoke(
            hobbit, [
                '--echo', 'startproject', '-n', 'haha', '-f', '-d', '.',
                '-p', '1024',
            ], obj={})
        assert result.exit_code == 0
        assert os.path.exists(os.path.join(
            os.getcwd(), 'app', 'models', '__init__.py'))
        assert 'example.py' not in result.output

    @chdir(wkdir)
    def test_startproject_cmd_example(self):
        assert os.getcwd() == self.wkdir
        runner = CliRunner()

        result = runner.invoke(
            hobbit, [
                '--echo', 'startproject', '-n', 'haha', '--example',
                '-p', '1024',
            ], obj={})
        # start + 27 files + 10 dir + 1 end + empty
        assert len(result.output.split('\n')) == 1 + 10 + 27 + 1 + 1
        assert result.exit_code == 0
        assert 'example.py' in result.output



import os
import random
import string

import click

from .handlers import echo
from .handlers.bootstrap import render_project
from hobbit_core import VERSION


@click.group()
@click.pass_context
def cli(ctx, force):
    pass


@cli.command()
@click.option('-n', '--name', help='Name of project.', required=True)
@click.option('-p', '--port', help='Port of web server.', required=True,
              type=click.IntRange(1024, 65535))
@click.option('-d', '--dist', type=click.Path(), required=False,
              help='Dir for new project.')
@click.option('-t', '--template', type=click.Choice(['shire']),
              default='shire', help='Template name.')
@click.option('-f', '--force', default=False, is_flag=True,
              help='Force render files, covered if file exist.')
@click.option('--example/--no-example', default=False,
              help='Generate api, test example files or not.')
@click.pass_context
def startproject(ctx, name, port, dist, template, force, example):
    """Create a new flask project, render from different template.

    Examples::

        hobbit --echo startproject -n demo -d /tmp/test --example -p 1024

    Other tips::

        hobbit --help
    """
    dist = os.getcwd() if dist is None else os.path.abspath(dist)
    ctx.obj['FORCE'] = force
    ctx.obj['EXAMPLE'] = example
    ctx.obj['JINJIA_CONTEXT'] = {
        'project_name': name,
        'port': port,
        'secret_key': ''.join(random.choice(
            string.ascii_letters) for i in range(38)),
        'version': '.'.join(str(i) for i in VERSION),
        'example': example,
    }

    echo('Start init a hobbit project `{}` to `{}`, use template {}',
         (name, dist, template))

    tpl_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'static', 'bootstrap', template)
    if not os.path.exists(tpl_path):
        raise click.UsageError(
            click.style('Tpl `{}` not exists.'.format(template), fg='red'))

    render_project(dist, tpl_path)

    echo('project `{}` render finished.', (name, ))


CMDS = [startproject]



from hobbit_core.flask_hobbit.schemas import ORMSchema, SchemaMixin

from marshmallow import fields

from .models import User


class UserSchema(ORMSchema, SchemaMixin):
    password = fields.Str(dump_only=True)

    class Meta:
        model = User



import os
from setuptools import setup, find_packages

import hobbit_core

SUFFIX = '.jinja2'
ROOT_PATH = os.path.split(os.path.abspath(os.path.join(__file__)))[0]
src_path = os.path.join(ROOT_PATH, 'hobbit_core')


def gen_data(data_root='hobbit/static/bootstrap'):
    """just for collect static files.
    """
    data = []
    for fn in os.listdir(os.path.join(src_path, data_root)):
        if os.path.isdir(os.path.join(src_path, data_root, fn)):
            data.extend(gen_data(os.path.join(data_root, fn)))
        if fn.endswith(SUFFIX):
            data.append(os.path.join(data_root, fn))
    return data


package_data = gen_data()
assert len(package_data) == 27, \
    'nums of tepl files error, {}'.format(len(package_data))


try:
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
except(OSError, ImportError):
    long_description = open('README.md').read()


setup(
    name='hobbit-core',
    version='.'.join(str(i) for i in hobbit_core.VERSION),
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*,'
                    ' !=3.4.*, !=3.5.*, <4',
    description='Hobbit - A flask project generator.',
    long_description=long_description,
    author='Legolas Bloom',
    author_email='zhanhsw@gmail.com',
    url='https://github.com/TTWShell/hobbit-core',
    packages=find_packages(),
    package_data={'': ['LICENSE'], 'hobbit_core': package_data},
    install_requires=[
        'Click==6.7',
        'Jinja2==2.10',

        'Flask==1.0.2',
        'flask-marshmallow==0.9.0',
        'Flask-Migrate==2.2.1',
        'flask-shell-ipython==0.3.1',
        'Flask-SQLAlchemy==2.3.2',
        'marshmallow-enum==1.4.1',
        'marshmallow-sqlalchemy==0.14.1',
        'webargs==4.0.0',
    ],
    entry_points={
        'console_scripts': 'hobbit = hobbit_core.hobbit:main'
    },
)



from flask import Blueprint, jsonify

from hobbit_core.flask_hobbit.utils import use_kwargs

from .schemas import UserSchema

bp = Blueprint('test', __name__)


@bp.route('/use_kwargs_with_partial/', methods=['POST'])
@use_kwargs(UserSchema(partial=True))
def use_kwargs_with_partial(**kwargs):
    return jsonify({k: v or None for k, v in kwargs.items()})


@bp.route('/use_kwargs_without_partial/', methods=['POST'])
@use_kwargs(UserSchema())
def use_kwargs_without_partial(**kwargs):
    return jsonify({k: v or None for k, v in kwargs.items()})



# -*- encoding: utf-8 -*-
import enum

from sqlalchemy import Integer, Column, ForeignKey, func, DateTime


class SurrogatePK(object):
    """A mixin that add ``id``、``created_at`` and ``updated_at`` columns
    to any declarative-mapped class.

    **id**: A surrogate integer 'primary key' column.

    **created_at**: Auto save ``datetime.now()`` when row created.

    **updated_at**: Auto save ``datetime.now()`` when row updated.
    """

    __table_args__ = {'extend_existing': True}

    id = Column(Integer, primary_key=True)
    created_at = Column(
        DateTime, nullable=False, server_default=func.now())
    updated_at = Column(
        DateTime, nullable=False, server_default=func.now(),
        onupdate=func.now())

    def __repr__(self):
        """You can set label property.

        Returns:
            str: ``<{classname}({pk}:{label!r})>``
        """
        return '<{classname}({pk}:{label!r})>'.format(
            classname=type(self).__name__, pk=self.id,
            label=getattr(self, 'label', ''))


def reference_col(tablename, nullable=False, pk_name='id', **kwargs):
    """Column that adds primary key foreign key reference.

    Args:
        tablename (str): Model.__table_name__.
        nullable (bool): Default is False.
        pk_name (str): Primary column's name.

    Others:

    See ``sqlalchemy.Column``

    Examples::

        from sqlalchemy.orm import relationship

        role_id = reference_col('role')
        role = relationship('Role', backref='users', cascade='all, delete')
    """

    return Column(
        ForeignKey('{0}.{1}'.format(tablename, pk_name)),
        nullable=nullable, **kwargs)


class EnumExt(enum.Enum):
    """ serialize/deserialize sqlalchemy enum field
    """
    @classmethod
    def strict_dump(cls, key, verbose=False):
        pos = 1 if verbose else 0
        return cls[key].value[pos]

    @classmethod
    def dump(cls, key, verbose=False):
        ret = {'key': cls[key].value[0], 'value': cls[key].value[1]}
        if verbose:
            ret.update({'label': key})
        return ret

    @classmethod
    def load(cls, val):
        pos = 1 if isinstance(val, str) else 0
        for elem in cls:
            if elem.value[pos] == val:
                return elem.name

    @classmethod
    def to_opts(cls, verbose=False):
        opts = []
        for elem in cls:
            opt = {'key': elem.value[0], 'value': elem.value[1]}
            if verbose:
                opt.update({'label': elem.name})
            opts.append(opt)
        return opts



"""
    flask_hobbit
    ~~~~~~~~~~~~

    Common utils for flask app.
"""

from flask import Flask


class HobbitManager(object):
    """Customizable utils management.
    """

    def __init__(self, app=None, **kwargs):
        """
        app: The Flask application instance.
        """
        self.app = app
        if app is not None:
            self.init_app(app, **kwargs)

    def init_app(self, app, **kwargs):
        """
        app: The Flask application instance.
        """
        if not isinstance(app, Flask):
            raise TypeError(
                'flask_hobbit.HobbitManager.init_app(): '
                'Parameter "app" is an instance of class "{}" '
                'instead of a subclass of class "flask.Flask".'.format(
                    app.__class__.__name__))

        # Bind Flask-Hobbit to app
        app.hobbit_manager = self



from hobbit_core.flask_hobbit.db import Column, SurrogatePK

from .exts import db


class User(SurrogatePK, db.Model):
    username = Column(db.String(50), nullable=True, unique=True)
    email = Column(db.String(50), nullable=True, unique=True)
    password = Column(db.String(255), nullable=False, server_default='')
